#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

/*
 *
 *  classes
 *
 */

typedef struct {
    int mentsu;
    int tahtsu;
    int toitsu;
    int float_suit_tiles;
    int dead_suit_tiles;
    int float_honor_tiles;
    int dead_honor_tiles;
} t_melds;


/*
 *
 *  function prototypes
 *
 */

void     try_and_print(int *tiles, int tile_count);
void    benchmark(int repeat_times);
int     compare_int(const int *a, const int *b);
int     random_haipai(int *buf, int limit);
void    print_usage();

int     tilearray2str(int *tiles, int tile_count, char *buf, int buf_size);
int     honorchr2tilenum(char c);
int     str2tilearray(char *str, int *buf, int buf_size);

int     shanten(int *tiles, int tile_count);
int     count_mentsu_shanten(int *tiles);
int     count_seventwins_shanten (int *tiles);
int     count_kokushi_shanten (int *tiles);
t_melds count_max_melds(int *tiles);
t_melds max_available_melds_in_suit(int *tiles);
t_melds max_available_melds_in_honor(int *tiles);
int     get_max_meld_comparer_in_suit_fast(int *tiles, int base, int repeater);


/*
 *
 *  global variables 
 *  (i know it's better to use const or enum, but i can't remember how to write. mandokuse.)
 *
 */

char *tile_chr[] = {
    "1c ", "2c ", "3c ", "4c ", "5c ", "6c ", "7c ", "8c ", "9c ",
    "1d ", "2d ", "3d ", "4d ", "5d ", "6d ", "7d ", "8d ", "9d ",
    "1b ", "2b ", "3b ", "4b ", "5b ", "6b ", "7b ", "8b ", "9b ",
    "T ", "N ", "S ", "P ", "W ", "G ", "R "};

char honor_chrs[] = {'t', 'n', 's', 'p', 'w', 'g', 'r'};

int opt_chin = 0;
int opt_time = 0;
t_melds last_max_melds;
int last_mentsu_shanten;
int last_seventwins_shanten;
int last_kokushi_shanten;

/*
 *
 *  HumanInterface Routines
 *
 */

int main(int argc,char *argv[])
{
    printf("welcome to shanten program!\n");
    printf("type q to quit, h to view help\n");
    while(1){
        printf("mj : ");
        char buff[255];
        fgets(buff, 255, stdin);
        int len = strlen(buff) - 1;
        buff[len] = 0; //cut off LF
        if( 0 == strcmp( buff, "random") || 0 == strcmp( buff, "r") || 0 == len){
            printf("randomhaipai!\n");
            int tiles[13];
            int res = random_haipai(tiles, 34);
            try_and_print(tiles, res);
        }
        else if( 0 == strcmp( buff, "chin") || 0 == strcmp( buff, "c") ){
            opt_chin = !opt_chin;
            printf("%s\n", opt_chin ? "enter the chin-ichi mode!" : "exit from chin-ichi mode!");
        }
        else if( 0 == strcmp( buff, "help") || 0 == strcmp( buff, "h") ){
            print_usage();
        }
        else if( 0 == strcmp( buff, "quit") || 0 == strcmp( buff, "q") ){
            printf("bye\n");
            break;
        }
        else if( 0 == strcmp( buff, "bench") || 0 == strcmp( buff, "b") ){
            printf("times? : ");
            fgets(buff, 255, stdin);
            len = strlen(buff) - 1;
            buff[len] = 0; //cut off LF
            int repeat_times = atoi(buff);
            benchmark(repeat_times);
        }
        else {
            int tiles[13];
            int res;
            if(res = str2tilearray(buff, tiles, 14)){
                try_and_print(tiles, res);
            }else{
                printf("what?\n");
            }
        }
    }
}

void try_and_print(int *tiles, int tile_count){
    int s = shanten(tiles, tile_count);
    t_melds m = last_max_melds;
    char str[255];
    tilearray2str(tiles, tile_count, str, 255);
    printf("--------\n");
    printf("checking for %d tiles...\n", tile_count);
    printf("sorted_tehai  : %s\n", str);
    printf("best mentsu : \n");
    printf("  mentsu | tahtsu | toitsu : %d | %d | %d\n", m.mentsu, m.tahtsu, m.toitsu);
    printf("  float suit  | dead suit  : %d | %d\n", m.float_suit_tiles, m.dead_suit_tiles);
    printf("  float honor | dead honot : %d | %d\n", m.float_honor_tiles, m.dead_honor_tiles);
    printf("each shanten:\n");
    printf("  mentsu | seventwins | kokushi : %d | %d | %d\n", last_mentsu_shanten, last_seventwins_shanten, last_kokushi_shanten);
    printf("shanten-suu : %d\n", s); 

}

void benchmark(int repeat_times){
    int i, sum = 0, mentsu_sum = 0, seventwins_sum = 0, kokushi_sum = 0;
    clock_t t1,t2;

    t1 = clock();

    for(i=0;i<repeat_times;i++){
        int tiles[13];
        int res;
        res = random_haipai(tiles, 34);
        sum += shanten(tiles,res);
        mentsu_sum += last_mentsu_shanten;
        seventwins_sum += last_seventwins_shanten;
        kokushi_sum += last_kokushi_shanten;
    }
    if(last_mentsu_shanten < 0) printf("out mentsn\n");
    if(last_seventwins_shanten < 0) printf("out seventwins\n");
    if(last_kokushi_shanten < 0) printf("out kokushi\n");
    t2 = clock();

    double tt = (double)(t2 - t1) / (double)CLOCKS_PER_SEC;
    printf("time = %fs for %d times roop. proc %f/s\n", tt, repeat_times, repeat_times/tt);
    printf("           avg = %f\n", (double)sum / (double)repeat_times);
    printf("mentsu     avg = %f\n", (double)mentsu_sum / (double)repeat_times);
    printf("seventwins avg = %f\n", (double)seventwins_sum / (double)repeat_times);
    printf("kokushi    avg = %f\n", (double)kokushi_sum / (double)repeat_times);
}

int compare_int(const int *a, const int *b){ return *a - *b; }

int random_haipai(int *buf, int limit){
    int deck[136];
    int i;
    for(i=0;i<136;i++) deck[i] = 1;
    for(i=0;i<13;i++) buf[i] = 0;
    for(i=0;i<13;){
        int rnd = rand() % (opt_chin ? 36 : 136);
        if(deck[rnd]){
            buf[i] = rnd / 4;
            deck[rnd]--;
            i++;
        }
    }
    qsort(buf, 13, sizeof(int), (int (*)(const void*, const void*))compare_int);
    return i;
}

void print_usage() {
    printf("random | r          : get random tiles and calculate shanten-suu of that hand.\n");
    printf("(enter w/o command) : one more alias of random command.\n");
    printf("bench | b           : try benchmark with random hand. enter the number of times to try.\n");
    printf("                      if you are in the chin-ichi mode, be careful to try with the big num.\n");
    printf("help | h            : view this help.\n");
    printf("chin | c            : toggle the chin-ichi mode on and off.\n");
    printf("                      in chin-ichi mode, use one suit only random hand, in random and bench command.\n");
    printf("quit | q            : quit this program.\n");
    printf("(your hand)         : type your hand, and you get an result of the calculation.\n");
    printf("                      here is markup rules;\n");
    printf("                      \n");
    printf("suit tiles:\n");
    printf("    basically, a tile will be marked-up by a pair of the rank number and suit character.\n");
    printf("        1b 2b 3b 5c 5c 5c 2d 5d 8d\n");
    printf("    here are suit characters:\n");
    printf("        b : bamboos\n");
    printf("        c : characters\n");
    printf("        d : dots\n");
    printf("    if the tiles are same suit, you can use block style. see below:\n");
    printf("        123b 555c 258d\n");
    printf("honor tiles:\n");
    printf("    just type an initial character.\n");
    printf("        T : Tong (east)\n");
    printf("        N : Nang (south)\n");
    printf("        S : Sha (west)\n");
    printf("        P : Pei (north)\n");
    printf("        W : White\n");
    printf("        G : Green\n");
    printf("        R : Red\n");
}


/*
 *
 *  Utility Routines
 *
 */

int tilearray2str(int *tiles, int tile_count, char *buf, int buf_size){
    int i=0,pos=0;
    for(i=0;i<tile_count;i++){
        char *chr = tile_chr[tiles[i]];
        strcpy(buf + pos, chr);
        pos += strlen(chr);
    }
    buf[pos] = 0;
    return 1;
}

int honorchr2tilenum(char c){
    int i=0;
    for(i=0;i<7;i++){
        if(honor_chrs[i] == c) return i + 27;
    }
    return 0;
}

int str2tilearray(char *str, int *buf, int buf_size){
    int pos, i=0, honor=0, suit_base = 0;
    for(pos = strlen(str) - 1; pos>=0; pos--){
        char c = str[pos];
        if('A' <= c && c <= 'Z'){
            c -= 0x20;
        }
        if(c == ' '){
            continue;
        }else if(c == 'c'){
            suit_base = -1;
        }else if(c == 'd'){
            suit_base = 8;
        }else if(c == 'b'){
            suit_base = 17;
        }else if('1' <= c && c <= '9'){
            if(!suit_base) return 0;
            buf[i++] = suit_base + (c - '0');
        }else if( honor = honorchr2tilenum(c) ){
            buf[i++] = honor;
        }else return 0;
    }
    qsort(buf, i, sizeof(int), (int (*)(const void*, const void*))compare_int);
    return i;
}


/*
 *
 *  Core Routines
 *
 */

int shanten(int *tiles, int tile_count){
    int i;
    int tile_abstract[34];
    for(i=0;i<34;i++)
        tile_abstract[i] = 0;
    for(i=0;i<tile_count;i++)
        tile_abstract[tiles[i]]++;
    int mentsu_shanten = count_mentsu_shanten(tile_abstract);
    int seventwins_shanten = count_seventwins_shanten(tile_abstract);
    int kokushi_shanten = count_kokushi_shanten(tile_abstract);

    last_mentsu_shanten = mentsu_shanten;
    last_seventwins_shanten = seventwins_shanten;
    last_kokushi_shanten = kokushi_shanten;

    int min;
    min = mentsu_shanten < seventwins_shanten ? mentsu_shanten : seventwins_shanten;
    min = min < kokushi_shanten ? min : kokushi_shanten;
    return min;
}

int count_mentsu_shanten(int *tiles){
    t_melds melds = count_max_melds(tiles);

    int shanten = 8;
    shanten -= melds.mentsu * 2;
    if(melds.tahtsu > 5 - melds.mentsu){
        shanten -= 5 - melds.mentsu;
    }else if(melds.tahtsu == 5 - melds.mentsu){
        shanten -= melds.tahtsu;
        if(!melds.toitsu) shanten++;
    } else {
        shanten -= melds.tahtsu;
    }

    int dead = melds.dead_suit_tiles + melds.dead_honor_tiles;
    if(melds.toitsu && melds.dead_suit_tiles) dead--;
    shanten = dead > shanten ? dead : shanten;
    return shanten;
}

int count_seventwins_shanten (int *tiles){
    int i, twins = 0, overplus = 0;
    for(i=0;i<34;i++){
        if(tiles[i] >= 2) twins++;
        if(tiles[i] > 2) overplus += tiles[i] - 2;
    }
    return (6 - twins) > overplus ? 6 - twins : overplus;
}

int count_kokushi_shanten (int *tiles) {
    int i,count=0, head=0;
    int targets[] = {0,8,9,17,18,26,27,28,29,30,31,32,33};
    for(i=0;i<13;i++){
        if(tiles[targets[i]]) count++;
        if(tiles[targets[i]] >= 2) head++;
    }
    return head ? 12 - count : 13 - count;
}

t_melds count_max_melds(int *tiles){
    t_melds total;
    total.mentsu            = 0;
    total.tahtsu            = 0;
    total.toitsu            = 0;
    total.float_suit_tiles  = 0;
    total.dead_suit_tiles   = 0;
    total.float_honor_tiles = 0;
    total.dead_honor_tiles  = 0;

    int i;
    t_melds melds_in_suit[4];

    melds_in_suit[0] = max_available_melds_in_suit(tiles     );
    melds_in_suit[1] = max_available_melds_in_suit(tiles + 9 );
    melds_in_suit[2] = max_available_melds_in_suit(tiles + 18);
    melds_in_suit[3] = max_available_melds_in_honor(tiles + 27);

    for(i=0;i<4;i++){
        total.mentsu            += melds_in_suit[i].mentsu;
        total.tahtsu            += melds_in_suit[i].tahtsu;
        total.toitsu            += melds_in_suit[i].toitsu;
        total.float_suit_tiles  += melds_in_suit[i].float_suit_tiles;
        total.dead_suit_tiles   += melds_in_suit[i].dead_suit_tiles;
        total.float_honor_tiles += melds_in_suit[i].float_honor_tiles;
        total.dead_honor_tiles  += melds_in_suit[i].dead_honor_tiles;
    }
    last_max_melds = total;
    return total;
}

t_melds max_available_melds_in_suit(int *tiles){
    int i, tile_count = 0;
    for(i=0;i<9;i++) tile_count += tiles[i];

    int max = get_max_meld_comparer_in_suit_fast(tiles, 0, 0);
    t_melds melds;
    melds.mentsu            = (max & 0xF000) >> 12;
    melds.tahtsu            = (max & 0x0F00) >> 8;
    melds.toitsu            = (max & 0x00F0) >> 4;
    melds.float_suit_tiles  = max & 0x000F;
    melds.dead_suit_tiles   = 0;
    melds.float_honor_tiles = 0;
    melds.dead_honor_tiles  = 0;


    melds.dead_suit_tiles = tile_count 
                          - melds.mentsu * 3
                          - melds.tahtsu * 2
                          - melds.float_suit_tiles;
    return melds;
}

t_melds max_available_melds_in_honor(int *tiles){
    t_melds melds;
    melds.mentsu            = 0;
    melds.tahtsu            = 0;
    melds.toitsu            = 0;
    melds.float_suit_tiles  = 0;
    melds.dead_suit_tiles   = 0;
    melds.float_honor_tiles = 0;
    melds.dead_honor_tiles  = 0;

    int i;
    for(i=0;i<7;i++){
        int cnt = tiles[i];
        if(cnt == 4){
            melds.mentsu++;
            melds.dead_honor_tiles++;
        }
        else if(cnt == 3){
            melds.mentsu++;
        }
        else if(cnt == 2){
            melds.tahtsu++;
            melds.toitsu++;
        }
        else if(cnt == 1){
            melds.float_honor_tiles++;
        }
    }
    return melds;
}

int get_max_meld_comparer_in_suit_fast(int *tiles, int base, int repeater){

    while(!(tiles[base])){
        base++;
        repeater = 0;
        if(base > 8) return 0;
    }

    int max = 0;
    int count;

    tiles[base]--;

    if (base < 7 && tiles[base + 2]){
        tiles[base + 2]--;
        count = get_max_meld_comparer_in_suit_fast(tiles, base, 1);
        count += 0x0100;
        max = max > count ? max : count;

        if (tiles[base + 1]){
            tiles[base + 1]--;
            count = get_max_meld_comparer_in_suit_fast(tiles, base, 1);
            count += 0x1000;
            max = max > count ? max : count;
            tiles[base + 1]++;
        }
        tiles[base + 2]++;
    }

    if (base < 8 && tiles[base + 1]){
        tiles[base + 1]--;
        count = get_max_meld_comparer_in_suit_fast(tiles, base, 1);
        count += 0x0100;
        max = max > count ? max : count;
        tiles[base + 1]++;
    }

    if(repeater){
        tiles[base]++;
        return max;
    }

    if(tiles[base] <= 2){
        count = get_max_meld_comparer_in_suit_fast(tiles, base, 1);
        count += 0x0001;
        max = max > count ? max : count;
    }

    if (tiles[base]){
        tiles[base]--;
        count = get_max_meld_comparer_in_suit_fast(tiles, base, 1);
        count += 0x0110;
        max = max > count ? max : count;

        if (tiles[base]){
            tiles[base]--;
            count = get_max_meld_comparer_in_suit_fast(tiles, base, 1);
            count += 0x1000;
            max = max > count ? max : count;

            if (tiles[base]){
                tiles[base]--;
                count = get_max_meld_comparer_in_suit_fast(tiles, base + 1, 0);
                count += 0x1000;
                max = max > count ? max : count;
                tiles[base]++;
            }
            tiles[base]++;
        }
        tiles[base]++;
    }

    tiles[base]++;
    return max;
}
